/*
 Arduino nRF24L01 RX Test
 Version: 0.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 19/09/2012
 
 Test the transmit/receive on an nRF24L01 with the Arduino
 
 VCC - 3.3V, CE - Pin 2, CSN - Pin 10
 SCK - Pin 13, MI - Pin 12, MO - Pin 11
 
 */

#include <avr/io.h>
#include <util/delay.h>
#include <SPI.h>
#include "mirf.h"
#include "nRF24L01.h"

// Defines for setting the MiRF registers for transmitting or receiving mode
#define TX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )

uint8_t data_in[1] = {0};
uint8_t data_out[1] = {2};

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    SPI.transfer(dataout[i]);
  }
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    datain[i] = SPI.transfer(0x00);
  }
}

// Initializes pins ans interrupt to communicate with the MiRF module
// Should be called in the early initializing phase at startup.
void mirf_init(void) {
  // Define CSN and CE as Output and set them to default
  pinMode (CSN, OUTPUT);
  pinMode (CE, OUTPUT);

  mirf_CE_lo;
  mirf_CSN_hi;
}

// Sets the important registers in the MiRF module and powers the module
// in receiving mode
void mirf_config(void) {
  // Set RF channel
  mirf_config_register(RF_CH, mirf_CH);

  // Set length of incoming payload 
  mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
  mirf_config_register(RX_PW_P1, mirf_PAYLOAD);

  // Set RADDR and TADDR
  mirf_write_register(RX_ADDR_P0, TADDR, 5);
  mirf_write_register(RX_ADDR_P1, RADDR, 5);
  mirf_write_register(TX_ADDR, TADDR, 5);

  // Enable RX_ADDR_P0 and RX_ADDR_P1 address matching since we also enable auto acknowledgement
  mirf_config_register(EN_RXADDR, (1<<ERX_P0 | 1<<ERX_P1));

  RX_POWERUP; // Power up in receiving mode
}

// Checks if data is available for reading
uint8_t mirf_data_ready(void) {
  mirf_CSN_lo;                                // Pull down chip select
  SPI.transfer(R_REGISTER | (REGISTER_MASK & STATUS));
  uint8_t status = SPI.transfer(0x00);        // Read status register
  mirf_CSN_hi;                                // Pull up chip select
  return status & (1<<RX_DR);
}

// Read the status register
uint8_t mirf_status(void) {
  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(R_REGISTER | (REGISTER_MASK & STATUS));
  uint8_t status = SPI.transfer(NOP); // Read status register
  mirf_CSN_hi; // Pull up chip select
  return status;
}

// Send data and wait for an ACK
uint8_t mirf_transmit_data(void) {
  TX_POWERUP; // Power up to transmitter mode

  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(FLUSH_TX);  // Write cmd to flush tx fifo
  mirf_CSN_hi; // Pull up chip select

  mirf_CSN_lo;  // Pull down chip select
  SPI.transfer(W_TX_PAYLOAD); // Write cmd to write payload
  spi_write_data(data_out, mirf_PAYLOAD); // Write payload
  mirf_CSN_hi; // Pull up chip select

  mirf_CE_hi; // Start transmission
  _delay_us(15);
  mirf_CE_lo;

  uint8_t waitcount = 0;
  while (!(mirf_status() & (1<<TX_DS))) { // Wait until we receive an ACK back
    _delay_us(250); // Wait the auto retransmit time
    if (mirf_status() & (1<<MAX_RT)) { // If maximum retries reached
      mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
      mirf_CE_hi;
      _delay_us(15);
      mirf_CE_lo;
    }
    waitcount++;
    if (waitcount >= 200) { //  If we haven't heard back in 50ms, exit
      mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
      return 0;
    }
  }
  mirf_config_register(STATUS,(1<<TX_DS)); // Reset status register

  return 1;
}

// Receive data
uint8_t mirf_receive_data(void) {
  RX_POWERUP;  
  mirf_CE_hi; // Start listening
  data_in[0] = 0; // Clear the data_in

  int waitcount = 0;
  while (!(mirf_status() & (1<<RX_DR))) { // Wait until data is ready to be read
    _delay_us(250);
    waitcount++;
    if (waitcount >= 200) { // If we haven't heard back in 50ms, exit
      mirf_CE_lo; // Stop listening
      return 0;
    }
  }
  mirf_CE_lo; // Stop listening

  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
  spi_read_data(data_in, mirf_PAYLOAD); // Read payload
  mirf_CSN_hi; // Pull up chip select
  mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register

  return 1;
}

// Write one byte into the given MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  SPI.transfer(value);
  mirf_CSN_hi;
}

// Reads an array of bytes from the given start position in the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
  spi_read_data(value, len);
  mirf_CSN_hi;
}

// Writes an array of bytes into inte the MiRF registers
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  spi_write_data(value, len);
  mirf_CSN_hi;
}

// Process the client check in
uint8_t process_check_in(void) {
  if (!mirf_receive_data()) { return 0; }
  if (!mirf_transmit_data()) { return 0; }

  return 1;
}

void setup() {
  Serial.begin(9600);
  SPI.begin();
  mirf_init();
  _delay_ms(50);
  mirf_config(); 
}
void loop() {

  while(1) {
    RX_POWERUP;
    mirf_CSN_lo; // Pull down chip select
    SPI.transfer(FLUSH_RX);  // Write cmd to flush rx fifo
    mirf_CSN_hi; // Pull up chip select
    mirf_CE_hi; // Start listening
    
    // Wait for incoming requests
    Serial.println("Waiting");
    while (!(mirf_status() & (1<<RX_DR))) { 
      _delay_us(250);
    }
    mirf_CE_lo; // Stop listening

    // Packet received, receive/read the client's request and transmit the action
    if (process_check_in()) {
      Serial.println("Packet received and reply sent");
      Serial.print("Data received was ");
      Serial.println(data_in[0], DEC);
    }
    else {
      Serial.println("Packet received - check in failed");
    }
  }
}


